/*
 *  This code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  This code is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "Pack.h"
#include <memory.h>
#include <math.h>


CPack::CPack()
{
	m_InputLBA = 0;		
	m_PackType = PackTypeNone;		
	m_NumButtons = 0;
	m_StreamID = 0;		
	m_FrameType = FrameTypeNone;	
	m_IFrameType = IFrameTypeNone;	
	m_IFrameTooFull = 0;	
	m_SRCH = 0;
	m_SRCL = 0;
	m_IFrameSize = 0;
}

CPack::~CPack()
{

}

void CPack::FindPackType(uchar* Buffer, ushort &VOBID, uchar &CELLID)
{
	// Get the SRC, all 6 bytes of it!
	m_SRCH = GetShort(Buffer,0x4);
	m_SRCL = GetInt(Buffer,0x6);

#if 0
	CompositeSRC CSRC;
	SrcToFrac(m_SRCL,m_SRCH,CSRC);
#endif

	// This function determines the current pack type. Here's what I use to identify packs,
	// this might not be robust.
	// video pack: start code at 000e = 1e0 
	// subpi pack: start code at 000e = 1bd. @016 header length. @016+L+1 stream ID == 20 or 21.
	// Audio pack: start code at 000e = 1bd. @016 header length. @016+L+1 stream ID == 80 or 81.
	// Nav Pack: private stream 2 code at 0026 = 1bf
	// Blank (arccos) packs: everything is 0.
	uint Temp = GetInt(Buffer,0x000e);
	if(Temp == 0x1e0)
	{
		// Video pack.
		m_StreamID = 0;		
		m_PackType = PackTypeVideo;
		return;
	}

	if(Temp == 0)
	{
		m_StreamID = 0;		
		m_PackType = PackTypeBlank;
		return;
	}

	if(Temp ==  0x1be)
	{
		m_StreamID = 0;		
		m_PackType = PackTypePadding;
		return;
	}

	if(Temp == 0x1bd)
	{
		// Subpic or Audio pack.
		uchar Length = Buffer[0x16];
		m_StreamID = Buffer[0x16+Length+1];
		if(m_StreamID >= 0x20 && m_StreamID <= 0x3f)
		{
			m_PackType = PackTypeSupPic;
			return;
		}
		if((m_StreamID >= 0x80 && m_StreamID <= 0x87) ||	// AC3
			(m_StreamID >= 0xa0 && m_StreamID <= 0xa7))		// LPCM
		{
			m_PackType = PackTypeAudio;
			return;
		}
	}

	// JEAN: Maybe I should also check for a private stream 2 at 0x400???
	if(GetInt(Buffer,0x0026) == 0x1bf && GetInt(Buffer,0x400) == 0x1bf)
	{
		// NAV PACK.
		m_PackType = PackTypeNavPack;
		VOBID = GetShort(Buffer,0x41f);
		CELLID = Buffer[0x422];
		m_NumButtons = Buffer[0x9e];
		return;
	}

	m_PackType = PackTypeNone;
	return;
}

int CPack::FindEndFrameOffset(uchar* Buffer)
{
	// This function looks for a start code in a video frame.
	// Look for the start code ("00000100" or 0x000001be or 0x000001b7 somewhere!)
	uchar* UCPtr;
	long* LPtr;
	// The header length is at 0x16 (1 byte). The video frame starts at 0x17+length.
	uint Length = (uint) (Buffer[0x16]);
	UCPtr = (Buffer+0x16+Length+1);

	LPtr = (long*) UCPtr;
	// Now look for "00000100" or 0x000001be or 0x000001b7 in the binary data.
	while(*LPtr != PICT_START_CODE && *LPtr != PADD_START_CODE && *LPtr != END_OF_SEQ_CODE && 
		UCPtr <= Buffer + SECTOR_SIZE - 4)
	{
		UCPtr++;
		LPtr = (long*) UCPtr;
	}
	if(*LPtr == END_OF_SEQ_CODE)
	{
		// Special case, the end of sequence is already there, we're cool.
		return 0;
	}
	return UCPtr-Buffer;
}

int CPack::FindClosedCaptions(uchar* Buffer)
{
	// This function looks for a closed caption header.
	// Look for a User Data start code (0x000001b2)
	uchar* UCPtr;
	long* LPtr;
	// The header length is at 0x16 (1 byte). The video frame starts at 0x17+length.
	uint Length = (uint) (Buffer[0x16]);
	UCPtr = (Buffer+0x16+Length+1);

	LPtr = (long*) UCPtr;
	// Now look for 0x000001b2 in the binary data.
	while(*LPtr != USER_START_CODE && UCPtr <= Buffer + SECTOR_SIZE - 4)
	{
		UCPtr++;
		LPtr = (long*) UCPtr;
	}
	if(UCPtr > Buffer + SECTOR_SIZE - 4)
		return 0; // Not found;
	// Found it! Now look whether the next data is a closed-caption header.
	LPtr++;
	if(*LPtr == CCAP_START_CODE)
		return 1;
	return 0;
}

void CPack::FindFrameType(uchar* Buffer, int* Offset, uchar* AspectRatio, int* CodedWidth, int* CodedHeight)
{
	// This function looks for a video frame type in a video frame.
	// First of, find a pic start code (00000100 somewhere!)
	uchar* UCPtr;
	long* LPtr;
	// The header length is at 0x16 (1 byte). The video frame starts at 0x17+length.
	uint Length = (uint) (Buffer[0x16]);
	UCPtr = (Buffer+0x16+Length+1);

	LPtr = (long*) UCPtr;
	// Now look for "00000100" in the binary data.
	while(*LPtr != PICT_START_CODE && UCPtr < Buffer + SECTOR_SIZE)
	{
		UCPtr++;
		LPtr = (long*) UCPtr;
	}
	if(Offset) *Offset = UCPtr-Buffer;
	// If we didn't find a picture start code, we're still in the previous picture.
	if(UCPtr >= Buffer + SECTOR_SIZE)
	{
		m_FrameType = FrameTypeNone;
		return;
	}
	// The frame type is bits 11-12-13 of the next 4 bytes (from the bottom for us). 
	UCPtr +=4 ;
	LPtr = (long*) UCPtr;
	uint Type = *LPtr & 0x00003800;	// Tricky mask! But it works.
	Type = Type >> 11;

	int CodedPictureWidth = 0;
	if(VideoFormat == FORMAT_UNKNOWN)
	{
		// We still don't know the video format. Let's determine it! Look for an MPEG sequence start code,
		UCPtr = (Buffer+0x16+Length+1);
		LPtr = (long*) UCPtr;
		// Now look for "000001b3" in the binary data.
		while(*LPtr != MPEG_START_CODE && UCPtr < Buffer + SECTOR_SIZE)
		{
			UCPtr++;
			LPtr = (long*) UCPtr;
		}
		// We found an MPEG sequence start code
		if(UCPtr < Buffer + SECTOR_SIZE)
		{
			UCPtr +=4 ;
			uint VideoAttributes = GetInt(UCPtr,0);
			// The bottom 12 bits of the three highest bytes are the vertical size.
			// http://dvd.sourceforge.net/dvdinfo/mpeghdrs.html
			VideoAttributes >>= 8;
			VideoAttributes &= 0x0fff;
			if(CodedHeight)
				*CodedHeight = VideoAttributes;
			if(VideoAttributes == 480 || VideoAttributes == 240)
				VideoFormat = FORMAT_NTSC;
			else
				VideoFormat = FORMAT_PAL;
			MESSAGE2("Found Video Format: %s\n",(VideoFormat == FORMAT_NTSC) ? "NTSC" : "PAL");
			MESSAGEF2("Format Video: %s\n",(VideoFormat == FORMAT_NTSC) ? "NTSC" : "PAL");
			
			// The aspect ratio are the high nibble of the 4th highest byte.
			if(AspectRatio)
			{
				*AspectRatio = uchar((GetInt(UCPtr,0) & 0x000000F0) >> 4);
			}

			if(CodedWidth)
			{
				VideoAttributes = GetInt(UCPtr,0);
				*CodedWidth = VideoAttributes >> 20;
			}
		}
	}

	m_FrameType = (FrameType) Type;
	return;
}


void CPack::TrimIFrame(uchar* Buf)
{
	// Here, we trim the last lba of an I-Frame, replacing the rest of the lba with a Padding
	// Stream.
	// First find offset to picture start code, or padding start code or end-of-sequence code.
	int PicStartOffset;
	uchar* UCPtr;
	long* LPtr;
	// The header length is at 0x16 (1 byte). The video frame starts at 0x17+length.
	uint Length = (uint) (Buf[0x16]);
	UCPtr = (Buf+0x16+Length+1);

	LPtr = (long*) UCPtr;
	// Now look for "00000100" or 0x000001be or 0x000001b7 in the binary data.
	while(*LPtr != PICT_START_CODE && *LPtr != PADD_START_CODE && *LPtr != END_OF_SEQ_CODE && 
		UCPtr < Buf + SECTOR_SIZE)
	{
		UCPtr++;
		LPtr = (long*) UCPtr;
	}
	// This should NOT happen! Don't do anything if it does.
	if(UCPtr >= Buf + SECTOR_SIZE)
	{
		ERROR1("\tFATAL ERROR! Didn't find end of I-frame! Can't put end_of_sequence or padding pack!\n");
		ERRORF1("\tERREUR! Impossible de trouver la fin de l'image. Impossible d'ajouter le code de\
 fin de squence\n");
		assert(UCPtr < Buf + SECTOR_SIZE);
		return;
	}
	// If we found 0x000001b7, the end_of_sequence code is already there!
	if(*LPtr == END_OF_SEQ_CODE)
		return;
	PicStartOffset = UCPtr-Buf;
	// Place end of sequence header (0x0lb7)
	PutInt(Buf,PicStartOffset,0x01b7);
	// Now, either we have enough space for a padding start code and its length, in which case we should
	// put one. Or we dont (in which case, we just fill with 0's until the end).
	int i;
	ushort SLength;
	if(PicStartOffset+4+4+2 <= SECTOR_SIZE)
	{
		// Place padding start code (0x0lbe)
		PutInt(Buf,PicStartOffset+4,0x01be);
		// Now, compute length and write it.
		SLength = SECTOR_SIZE - PicStartOffset - 6 - 4;
		assert(SLength < SECTOR_SIZE);
		PutShort(Buf,PicStartOffset+8,SLength);

		// Then fill the first with 0xff.
		for(i=PicStartOffset+10;i<SECTOR_SIZE;i++)
			Buf[i] = 0xff;

		// Now modify the frame length at offset 0x12. The length is exactly:
		// Index of last good byte of data + 1 - 0x14 (20). For a full frame: 2048-20 = 2028.
		// Apparently, the end_of_sequence code must be included as good bytes of data.
		SLength = PicStartOffset + 3 - 23 + 4;
		PutShort(Buf,0x12,SLength);
	}
	else
	{
		// Fill with 0's until the end of the pack.
		for(i=PicStartOffset+4;i<SECTOR_SIZE;i++)
			Buf[i] = 0x00;
		SLength = PicStartOffset + 4 - 20;
		PutShort(Buf,0x12,SLength);
	}
}

void CPack::CloseGOP(uchar* Buffer)
{
	// Here, Buf points to the first lba of an I-frame.
	// This function will set the Closed GOP flag, and reset the Broken GOP flag.
	// First of, find the start of the video frame.
	uchar* UCPtr;
	long* LPtr;
	// The header length is at 0x16 (1 byte). The video frame starts at 0x17+length.
	uint Length = (uint) (Buffer[0x16]);
	UCPtr = (Buffer+0x16+Length+1);

	LPtr = (long*) UCPtr;
	// Now look for "000001b8" in the binary data. That's the GOP start code.
	while(*LPtr != GOP__START_CODE && UCPtr < Buffer + SECTOR_SIZE)
	{
		UCPtr++;
		LPtr = (long*) UCPtr;
	}
	// Didn't find a GOP start code?!
	if(UCPtr >= Buffer + SECTOR_SIZE)
	{
		assert(0);
		ERROR1("WARNING: Didn't find GOP start code, can't set Closed GOP flag\n");
		ERRORF1("ATTENTION: Pas de code \"Start GOP\", impossible de corriger le bit \"Closed GOP\"\n");
		return ;
	}

	// Now set the closed GOP flag and reset the broken GOP flag. 
	UCPtr +=4 ;
	uint Flag = GetInt(UCPtr,0);
	Flag |= 0x40;
//	Flag &= ~0x40;	// mpucoder insistes that the closed GOP flag should be 0 for a single I-frame.
	Flag &= ~0x20;
	PutInt(UCPtr,0,Flag);

	// Now let's set the temporal sequence number to 0.
	// First, let's find the picture start code.
	UCPtr = (Buffer+0x16+Length+1);
	LPtr = (long*) UCPtr;
	// Now look for "00000100" in the binary data. That's the picture start code.
	while(*LPtr != PICT_START_CODE && UCPtr < Buffer + SECTOR_SIZE)
	{
		UCPtr++;
		LPtr = (long*) UCPtr;
	}
	// Didn't find a picture start code?!
	if(UCPtr >= Buffer + SECTOR_SIZE)
	{
		assert(0);
		ERROR1("WARNING: Didn't find picture start code, can't adjust the temporal sequence number\n");
		ERRORF1("ATTENTION: Pas de \"picture start code\", impossible d'ajuster le code de squence temporelle\n");
		return ;
	}
	// Get picture header:
	UCPtr +=4 ;
	uint Header = GetInt(UCPtr,0);
	// The temporal sequence number occupies the 10 highest bits.
	Header &= 0x003FFFFF;
	PutInt(UCPtr,0,Header);
}
	
int CPack::FindSubpicLengthAndOffset(uchar* Buffer, ushort &Length, ushort &Offset, ushort &TrueOffset)
{
	// Get to the start of the subpic stream. len is the length of the header data.
	uint len = (uint) (Buffer[0x16]);
	// The first value in the subpic stream is the length, the second one the offset to the commands.
	Length = GetShort(Buffer,0x18+len);
	TrueOffset = GetShort(Buffer,0x1a+len);
	// Length and Offset are relative to the start of the subpic data, at 0x18+len
	Length += 0x18+len;
	Offset = TrueOffset + 0x18+len;
	return 0;
}

int CPack::RemoveSubpicStop(uchar* Buffer, uchar* Buffer2, ushort TrueOffset, ushort Offset)
{
	int i;
	uchar* FirstBuf = Buffer;
	// Offset points exactly to the start of the command area, which starts with two 0 bytes which
	// represent the time stamp for the first set of commands. This time has to be 0.
	assert(Buffer[Offset] == 0 && Buffer[Offset+1] == 0);
	// I'm not really checking for this. It might come back to bite me!
	assert(Offset+3 < SECTOR_SIZE);
	i = Offset+2;
	// i now points to the first link. 
	uchar Link[2];
	Link[0] = Buffer[i];
	Link[1] = Buffer[i+1];	// JEAN: Could check whether this points to itself.
	ushort PointTo = TrueOffset + (ushort) i - Offset - 2;
	if(PointTo == GetShort(Link,0))
		return 0;
	// Now look for another instance of that link. The link can either point to itself or another set
	// of commands. If there's a stop command, it has to be in another set, timed later in the cell, or
	// the subpics would stop right away! That second set, and there's usually no more than 2, then has
	// to point to itself. Hence the two identical links.
	i+=2;
	for(;i<SECTOR_SIZE-1;i++)
	{
		if(Buffer[i] == Link[0] && Buffer[i+1] == Link[1])
			break;
	}
	if(i>=SECTOR_SIZE-1)
	{
		// Didn't find a second link, maybe it's in the second buffer (buffer2). 
		if(Buffer2 == 0)
			return -1;		// Nope... simply not there.
		Buffer = Buffer2;
		for(i=0x18;i<SECTOR_SIZE-1;i++)
		{
			if(Buffer[i] == Link[0] && Buffer[i+1] == Link[1])
				break;
		}
		if(i==SECTOR_SIZE-1)
			return -1;		// Not in second buffer either.
		TrueOffset += (SECTOR_SIZE - 0x18);	// awful bookkeeping.
	}
	// As a safety check, this second link should point to itself.
	PointTo = TrueOffset + (ushort) i - Offset - 2; 
	assert(PointTo == GetShort(Link,0));
	// Found the second instance of the link! The 2 bytes left of it are the time at which the commands
	// are supposed to execute. Let's set that to zero.
	// However, it's possible that one or both of these guys is/are in the previous pack!
	if(i-1 >= 0x18)
		Buffer[i-1] = 0;
	else
		FirstBuf[SECTOR_SIZE-1]=0;
	if(i-2 >= 0x18)
		Buffer[i-2] = 0;
	else
		FirstBuf[SECTOR_SIZE-(0x18-i+2)]=0;
	// The next byte is the command, probably a stop ('02'). Set it to FF, which means "end 
	// of command area".
	assert(Buffer[i+2] == 2 || Buffer[i+2] == 0xff);
	if(Buffer[i+2] != 2 && Buffer[i+2] != 0xff)
	{
		assert(0);
		MESSAGE1("Something is strange, found a timed command that isn't a stop...\n");
		MESSAGEF1("Etrange: une commande avec delai a t trouve, mais ca n'est pas un stop!...\n");
		return 0;
	}
	if(Buffer[i+2] == 2)
	{
		MESSAGE2("Fixing subpic timed stop command. Stream 0x%x\n",m_StreamID);
		MESSAGEF2("Enlvement de la commande de stop avec dlai. Flux 0x%x\n",m_StreamID);
	}
	Buffer[i+2]=0xff;
	return 0;
}

void CPack::FindMaxButtonSize(uchar* Buffer, ushort &MaxSizeX, ushort &MaxSizeY, int &AllNops,
							  int &ForcedAction, int &AnyNext)
{
	MaxSizeX = 0;
	MaxSizeY = 0;
	int i;
	AllNops = 1;
	ForcedAction = 0;
	AnyNext = 0;
	// Read number of buttons:
	uchar NumButtons = Buffer[0x9e];
	long Add = 0xbb;
	for(i=0;i<NumButtons;i++,Add+=18)
	{
		ushort StartX = ((ushort)(Buffer[Add] & 0x3f)) << 4; StartX |= ((ushort)(Buffer[Add+1] & 0xf0)) >> 4;
		ushort EndX = ((ushort)(Buffer[Add+1] & 0x03)) << 8; EndX |= (ushort)(Buffer[Add+2]);
		assert(EndX >= StartX);
		EndX -= StartX; if(MaxSizeX < EndX) MaxSizeX = EndX;

		ushort StartY = ((ushort)(Buffer[Add+3] & 0x3f)) << 4; StartY |= ((ushort)(Buffer[Add+4] & 0xf0)) >> 4;
		ushort EndY = ((ushort)(Buffer[Add+4] & 0x03)) << 8; EndY |= (ushort)(Buffer[Add+5]);
		assert(EndY >= StartY);
		EndY -= StartY; if(MaxSizeY < EndY) MaxSizeY = EndY;

		// See if all the buttons are either (NOP or LinkNextPG or LinkNextCell).
		if((GetInt(Buffer,Add+0x0a) != 0 || GetInt(Buffer,Add+0x0e) != 0) && 
			(GetInt(Buffer,Add+0x0a) != 0x20010000 || GetInt(Buffer,Add+0x0e) != 0x00000006) &&
			(GetInt(Buffer,Add+0x0a) != 0x20010000 || GetInt(Buffer,Add+0x0e) != 0x00000002))
			AllNops = 0;
		// See if any of the commands uses a LinkNextPG or LinkNextCell
		// Links start either with 0x2001 in the 2 first bytes, or with 8-9-a-b-c-d
		// in the first nibble of the first byte. The last nibble is 6 for NextPG and
		// 2 for NextCell. This checks for compounded commands.
		if(((GetInt(Buffer,Add+0x0e) & 0x0000000F) == 0x06 ||
			(GetInt(Buffer,Add+0x0e) & 0x0000000F) == 0x02) &&
			(
				(GetInt(Buffer,Add+0x0a) & 0xFFFF0000) == 0x20010000 ||
				(GetInt(Buffer,Add+0x0a) & 0xF0000000) > 0x80000000
			)
			)
			AnyNext = 1;

		if(Buffer[0xa2] != 0)
			ForcedAction = 1;
	}
}

void CPack::GetButtonInfo(uchar* Buffer, ButtonInfo* AllInfos)
{
	int i;
	// Read number of buttons:
	uchar NumButtons = Buffer[0x9e];
	long Add = 0xbb;
	for(i=0;i<NumButtons;i++,Add+=18)
	{
		ushort StartX = ((ushort)(Buffer[Add] & 0x3f)) << 4; StartX |= ((ushort)(Buffer[Add+1] & 0xf0)) >> 4;
		ushort EndX = ((ushort)(Buffer[Add+1] & 0x03)) << 8; EndX |= (ushort)(Buffer[Add+2]);
		AllInfos[i].m_startx=StartX;
		assert(EndX >= StartX);
		AllInfos[i].m_sizex=EndX-StartX;

		ushort StartY = ((ushort)(Buffer[Add+3] & 0x3f)) << 4; StartY |= ((ushort)(Buffer[Add+4] & 0xf0)) >> 4;
		ushort EndY = ((ushort)(Buffer[Add+4] & 0x03)) << 8; EndY |= (ushort)(Buffer[Add+5]);
		AllInfos[i].m_starty=StartY;
		assert(EndY >= StartY);
		AllInfos[i].m_sizey=EndY-StartY;

		AllInfos[i].m_auto = (Buffer[Add+3] >> 6);
		AllInfos[i].m_up = (Buffer[Add+6] & 0x3f);
		AllInfos[i].m_down = (Buffer[Add+7] & 0x3f);
		AllInfos[i].m_left = (Buffer[Add+8] & 0x3f);
		AllInfos[i].m_right = (Buffer[Add+9] & 0x3f);


	}
}


uchar* CPack::GetColorScheme(uchar* Buffer, int Scheme)
{
	if(m_PackType != PackTypeNavPack) return 0;
	return Buffer+0xa3+Scheme*8;
}


 // ________________________________________________________________________________________


CDVDTime::CDVDTime()
{
	m_hours = m_minutes = m_seconds = m_frames = 0;
	m_format = FORMAT_NTSC;
}

CDVDTime::~CDVDTime()
{

}

CDVDTime::CDVDTime(int PTSDuration)
{
	m_format = VideoFormat;
	// Find number of frames.
	m_frames = PTSDuration / ((m_format == FORMAT_PAL) ? 3600 : 3003);
	int Fps = (m_format == FORMAT_PAL) ? 25 : 30;
	m_seconds = m_frames / Fps;
	m_frames = m_frames % Fps;
	m_minutes = m_seconds / 60; 
	m_seconds %= 60;
	m_hours = m_minutes / 60; 
	m_minutes %= 60;
}

CDVDTime &CDVDTime::operator += ( CDVDTime &RHS )
{
	m_hours += RHS.m_hours;
	m_minutes += RHS.m_minutes;
	m_seconds += RHS.m_seconds;
	m_frames += RHS.m_frames;

	int Fps = (m_format == FORMAT_PAL) ? 25 : 30;

	m_seconds += m_frames / Fps;
	m_frames = m_frames % Fps;
	m_minutes += m_seconds / 60; 
	m_seconds %= 60;
	m_hours += m_minutes / 60; 
	m_minutes %= 60;
	return *this;
}

CDVDTime &CDVDTime::operator -= ( CDVDTime &RHS )
{
	m_hours -= RHS.m_hours;
	m_minutes -= RHS.m_minutes;
	m_seconds -= RHS.m_seconds;
	m_frames -= RHS.m_frames;

	int Fps = (m_format == FORMAT_PAL) ? 25 : 30;

	// Tricky tricky. Easiest solution is to use the floor function. 
	m_seconds += (int) floor((float)m_frames / Fps);
	m_frames -= Fps * (int) floor((float)m_frames / Fps);;
	m_minutes += (int) floor((float)m_seconds / 60); 
	m_seconds -=  60 * (int) floor((float)m_seconds / 60);
	m_hours += (int) floor((float)m_minutes / 60); 
	m_minutes -= 60 * (int) floor((float)m_minutes / 60);
	assert(m_frames >= 0);
	assert(m_seconds >= 0);
	assert(m_minutes >= 0);
	assert(m_hours >= 0);

	return *this;
}

bool CDVDTime::operator < ( CDVDTime &RHS )
{
	bool Ret = false;
	if(m_hours < RHS.m_hours) return true;
	if(m_hours > RHS.m_hours) return false;

	// At this point, hours are equal.
	if(m_minutes < RHS.m_minutes) return true;
	if(m_minutes > RHS.m_minutes) return false;

	// At this point, hours and minutes are equal.
	if(m_seconds < RHS.m_seconds) return true;
	if(m_seconds > RHS.m_seconds) return false;

	// At this point, hours minutes and seconds are equal.
	if(m_frames < RHS.m_frames) return true;
	else  return false;
}

float CDVDTime::ConvertToSeconds()
{
	return m_hours*3600 + m_minutes*60 + m_seconds + (float) m_frames / 
		((m_format == FORMAT_PAL) ? 25 : 30);
}

void CDVDTime::ReadFromBuffer(uchar* Buffer)
{
	// The format for how the elapsed time (and cell and PGC times) are saved is really strange.
	// hours minutes seconds and frames take 1 byte each. The 4 highest bits code the first decimal
	// digit, the 4 lowest ones the last decimal digit! For example, 13 is codec as 0x13
	// In addition, bits 7 and 6 have info on the frame rate. 0x40 for PAL, 0xc0 for NTSC.
	m_hours = m_minutes = m_seconds = m_frames = 0;
	uint PBTime = GetInt(Buffer,0);
	uchar Temp = (PBTime & 0xFF000000)>>24; Temp = Temp%16+10*(Temp/16); m_hours += Temp;
	Temp = (PBTime & 0x00FF0000)>>16; Temp = Temp%16+10*(Temp/16); m_minutes += Temp;
	Temp = (PBTime & 0x0000FF00)>>8; Temp = Temp%16+10*(Temp/16); m_seconds += Temp;
	Temp = (PBTime & 0x0000003F); Temp = Temp%16+10*(Temp/16); m_frames += Temp;

	m_format = ((PBTime & 0x000000c0) == 0x00000040) ? FORMAT_PAL : FORMAT_NTSC;
}

void CDVDTime::WriteToBuffer(uchar* Buffer)
{
	uint PBTime = 0;
	uchar Temp = 16*(m_hours/10)+(m_hours%10);PBTime += (uint)Temp << 24;
	Temp = 16*(m_minutes/10)+(m_minutes%10);PBTime += (uint)Temp << 16;
	Temp = 16*(m_seconds/10)+(m_seconds%10);PBTime += (uint)Temp << 8;
	Temp = 16*(m_frames/10)+(m_frames%10);PBTime += (uint)Temp;
	PBTime |= ((m_format == FORMAT_PAL) ? 0x00000040 : 0x000000c0);
	PutInt(Buffer,0,PBTime);
}

bool CDVDTime::IsZero()
{
	return (m_hours == 0 && m_minutes == 0 && m_seconds == 0 && m_frames == 0);
}

CompositeSRC ParseSRC(uchar* NavPack)
{
	CompositeSRC Ret;
	ushort SRCH = GetShort(NavPack,0x4);
	uint SRCL = GetInt(NavPack,0x6);
	SrcToFrac(SRCL, SRCH, Ret);
	return Ret;
}

void WriteSRC(uchar* NavPack, CompositeSRC SRC) 
{
	ushort SRCH;
	uint SRCL;
	FracToSrc(SRC, SRCL, SRCH);
	PutShort(NavPack,0x4,SRCH);
	PutInt(NavPack,0x6,SRCL);
}

void FracToSrc(CompositeSRC CSRC, uint &SRCL, ushort &SRCH)
{

	// Put the guard bits;
	SRCL = 0x04000401;
	SRCL |= (uint)CSRC.m_SRCFrac << 1;
	SRCL |= (CSRC.m_SRCInt & 0x7FFF) << 11;
	SRCL |= (CSRC.m_SRCInt & ~0x7FFF) << 12;

	SRCH = 0x4400;
	SRCH |= (CSRC.m_SRCInt >> 20) & 0x03FF;
	SRCH |= (CSRC.m_SRCInt >> 19) & 0x3800;
	SRCH |= (uint)CSRC.m_SRCTop << 13;
}

void SrcToFrac(uint SRCL, ushort SRCH, CompositeSRC &CSRC)
{
	// Goes from the 6 bytes representing the SRC in the VOB, to a int/frac representation.
	// Get the frac first: It's 9 bits wide starting at bit 1.	
	CSRC.m_SRCFrac = ((ushort) SRCL >> 1) & 0x01FF;
	// The int part is a bit trickier.
	CSRC.m_SRCInt = (uint)(SRCL >> 11) & 0x7FFF;		// Starts at bit 11 of SRCL.
	CSRC.m_SRCInt |= (uint)(SRCL & 0xF8000000) >> 12;	 
	CSRC.m_SRCInt |= ((uint)(SRCH & 0x03FF)) << 20;
	CSRC.m_SRCInt |= ((uint)(SRCH & 0x3800)) << 19;

	CSRC.m_SRCTop = (uchar)((SRCH & 0x2000) >> 13);
}

void AddSRC(CompositeSRC &SRC, CompositeSRC SRC2)
{
	SRC.m_SRCFrac += SRC2.m_SRCFrac;
	SRC.m_SRCInt += SRC2.m_SRCInt;
	if(SRC.m_SRCFrac >= 300)
	{
		SRC.m_SRCFrac -= 300;
		SRC.m_SRCInt ++;
	}
}

void IncSRC(CompositeSRC &SRC)
{
	SRC.m_SRCFrac += 86;
	if(SRC.m_SRCFrac >= 300)
	{
		SRC.m_SRCFrac -= 300;
		SRC.m_SRCInt ++;
	}
	SRC.m_SRCInt += 146;
}

bool SRC1GESRC2(CompositeSRC SRC1, CompositeSRC SRC2)
{
	return (SRC1.m_SRCInt > SRC2.m_SRCInt || (
		SRC1.m_SRCInt == SRC2.m_SRCInt && SRC1.m_SRCFrac >= SRC2.m_SRCFrac));
}

uint ParsePTS(uchar* PTSIn)
{
	// Parses the 5 bytes PTS and returns it as a 4 byte value.
	uint PTSOut = 0;
	PTSOut |= ((uint)PTSIn[4] >> 1);			// bits 0-6
	PTSOut |= ((uint)PTSIn[3] << 7);			// bits 7-14
	PTSOut |= ((uint)(PTSIn[2] & 0xFE) << 14);	// bits 15-21
	PTSOut |= ((uint)(PTSIn[1]) << 22);			// bits 22-29
	PTSOut |= ((uint)(PTSIn[0] & 0x0E) << 29);	// bits 30-31
	return PTSOut;
}

void BuildPTS(uchar* PTSOut, uint PTSIn, uchar PTSPreset)
{
	PTSOut[4] = (uchar) (PTSIn << 1) | 0x01;
	PTSOut[3] = (uchar) (PTSIn >> 7);
	PTSOut[2] = (uchar) (PTSIn >> 14) | 0x01;;
	PTSOut[1] = (uchar) (PTSIn >> 22);
	PTSOut[0] = (uchar) (PTSIn >> 29) | 0x01;;
	PTSOut[0] |= PTSPreset;
}

void CNavPack::GetVCID(ushort &VOBID, uchar &CELLID)
{
	VOBID = GetShort(m_Buffer,0x41f);
	CELLID = m_Buffer[0x422];
}

uint CNavPack::GetOffsetToNextNavPack()
{
	return GetInt(m_Buffer,0x040f);
}
uchar CNavPack::GetNumButtons()
{
	uchar NumGroups = (m_Buffer[0x9b] >> 4) & 0x03;
	return NumGroups ? m_Buffer[0x9e] : 0;	
}

ushort CNavPack::GetHighlight()
{
	return GetShort(m_Buffer,0x8d);
}

void CNavPack::GetButtonCommand(uchar Button, uchar* Command)
{
	memcpy(Command,m_Buffer+0xc5+Button*18,8);
}

void CNavPack::GetButtonSize(uchar Button, ushort &width, ushort &height)
{
	long Add = 0xbb + 18*Button;
	ushort StartX = ((ushort)(m_Buffer[Add] & 0x3f)) << 4; 
	StartX |= ((ushort)(m_Buffer[Add+1] & 0xf0)) >> 4;
	ushort EndX = ((ushort)(m_Buffer[Add+1] & 0x03)) << 8; 
	EndX |= (ushort)(m_Buffer[Add+2]);

	width = EndX-StartX;

	ushort StartY = ((ushort)(m_Buffer[Add+3] & 0x3f)) << 4; 
	StartY |= ((ushort)(m_Buffer[Add+4] & 0xf0)) >> 4;
	ushort EndY = ((ushort)(m_Buffer[Add+4] & 0x03)) << 8; 
	EndY |= (ushort)(m_Buffer[Add+5]);

	height = EndY-StartY;
}

