#ifndef VTS_CLEAN_H
#define VTS_CLEAN_H

#include "includes.h"
#include "IFOUtils.h"

// THis enum indicates what we're currently doing:
enum ProcessPhase
{
	Processing_Idle,		// Nothing
	Processing_VOB,			// Processing VOB files.
	Processing_IFO,			// Processing IFO files
	Writing_IFO,			// Writing back IFO file.
	Processing_Done,		// Done with processing.
	Processing_Aborted,		// Processing was aborted.
};

class CVTSClean
{
public:
	CVTSClean();
	~CVTSClean();

	// Opens this file for processing. It will open the first file of the domain whatever you pass.
	virtual int		OpenFile(const char* FileName);

	// Called when a new cell is encountered. If you override, make sure you call the base class!
	virtual int		OnNewCell(CIFOLoader* IFO); 

	virtual void	RestoreBackup();	// Restore backup files.
	virtual void	DeleteBackup();		// Remove backup files and backup directory.
	virtual void	Process();			// Start processing.

	virtual void	ProcessFullDVD();	// This calls Process() on all the files!

	void			EraseLog();			// Erases the log file.

	int				m_RemoveUnrefed;	// If true, unreferenced cells will be removed from output.
	int				m_RemoveBlank;		// If true, blank packs are removed. If not, they're replaced with stuffing packs.
	int				m_InPlaceProcess;	// If true, process VOBs from original. If not, process from backup.
	int				m_RemapVCID;		// If true, the VCIDs are remapped so the CID are consecutive.
	int				m_RemapVOBID;		// If true the VOBID are remapped to be consecutive as well.
	int				m_ResetILVUPtrs;	// If true angle pointers are reset to 0.
	int				m_WaitForFile;		// If true, fixvts waits for the file to become available for processing.
	int				m_PutBackupAbove;	// If true, the backup directory is at the same level as VIDEO_TS
	int				m_FixPGCLBAPtrs;	// If true, the LBA pointers in the PGCs are fixed if incorrect.

	char			m_Version[512];		// Version string.
	bool			m_AbortNow;			// Set this to true to stop processing.
	bool			m_Running;			// Indicates we're currently processing.
	ProcessPhase	m_Phase;			// indicates what we're currently doing.
	bool			m_Pause;			// Set to true to pause processing.
	char			m_FileName[1024];	// Name of file currently being processed.

	int				m_NumFwdPtrFixed;	// How many forward pointers we fixed.
	int				m_NumPacksRemoved;	// How many packs were removed.
	int				m_NumPacksRemapped;	// How many navpacks had their VCID changed.
	int				m_NumILVUPtrReset;	// How many navpacks had their ILVU ptrs reset.
	static FILE*	m_Outlog;			// The output log FD.
	char			m_LogPath[1024];	// The output log path.


protected:
	// Helper function to open files.
	char*			GetBackupName(char* FileName);
	int				ReadOpenThisFile(char* FileName, bool FromBackup);
	int				WriteOpenThisFile(char* FileName);
	void			RestoreThisFile(char* FileName);

	// Opens IFO file and return CIFOLoader class.
	CIFOLoader*		OpenIFO(char* FileName);
	// This processes 1 pack.
	int				ProcessPack(CIFOLoader* IFO);
	// This processes the IFO files.
	int				ProcessIFOs(CIFOLoader* IFO);

	// Finds the next vob to be processed.
	bool			FindNextVOB(char* FileName, bool CheckEmpty = true);

	// This counts the number of input lbas, over all input files.
	int				CountInputLBAs();

	// Resets before processing...
	void			Reset();

	// Checks whether the VCID is in our map.
	bool			AlreadyInVCIDMap(ushort VOBID, uchar CELLID);
	// Adds the VCID to the VCID mapping table. Assigns it a new CELLID based on m_NextCID. Returns
	// the new VCID as VOBID*256+CELLID.
	uint			AddToVCIDMap(ushort VOBID, uchar CELLID);
	uint			AddToVCIDMap2(ushort VOBID, uchar CELLID);
	// Retrieve the mapped CELLID for a given input VCID. The VOBID does not change in the map. Returns
	// true if the new CID is different than the original one.
	bool			GetMapping(ushort VOBID, uchar CELLID, ushort &NewVOBID, uchar &NewCELLID);

	// Helper function that resets angle pointers in navpacks. Returns true if the pack needs
	// to be written (one of the angles was set).
	bool			ResetILVUPtrs();

	// This goes through all the vobs, removes the empty ones and renames the other ones...
	void			RemoveEmptyVobs();
	void			UpdateIFOForEmptyVob();

	void			UpdateLastNavPack(int CumulNavPackLBA, int ForwardPointer);

	int				m_NumInputLBA;			// Number of LBAs in current input file.
	int				m_CumulNumInputLBA;		// Number of LBAs, counting all input files.
	int				m_InPackIndex;			// Current input LBA in current input file.
	int				m_OutPackIndex ;		// Current output LBA in current output file.
	int				m_CumulInPackIndex ;	// Cumulative input LBA (over all input files)
	int				m_CumulOutPackIndex ;	// Cumulative output LBA (over all input files).
	int				m_LastNavPack ;			// Non cumulative LBA or last nav pack encountered.
	int				m_CumulLastNavPack ;	// Cumulative LBA or last nav pack encountered.
	int				m_LastNavPackFwdPtr;	// Forward pointer read in preview navpack.
	ushort			m_VOBID ;				// Current VOBID.
	uchar			m_CELLID ;				// Current CELLID.
	uchar			m_NewCELLID ;			// New (remapped) current CELLID.
	uchar			m_Data[SECTOR_SIZE];	// Temporary buffers.
	uchar			m_Prev[SECTOR_SIZE];
	int				*m_InNavPacks ;			// Array of input navpack LBAs. Used to adjust IFO's VTS_VOBU_ADMAT
	int				*m_OutNavPacks ;		// Array of output navpack LBAs. Used to adjust IFO's VTS_VOBU_ADMAT
	int				m_NavPackCount ;		// How many navpacks did we encounter.
	bool			m_SkipThisCell ;		// Indicates this cell should be skipped.
	int				m_NumCells ;			// Number of VCIDs we found.
	FILE*			m_InFileFD ;			// Input file FD pointer.
	FILE*			m_OutFileFD ;			// Output file FD pointer.
	VTSM_C_ADTItem	*m_CellInfos;			// Array of cell infos, used to update VTS_C_ADT.
	uchar			m_StuffingPack[SECTOR_SIZE];	// Stuffing pack, used to replace blank packs.

	bool			m_MenuDomain;			// True if processing menu domain (vts_xx_0.vob or video_ts.vob).

	int				m_VCIDMapLength;		// Length of our VCID mapping table.
	VCIDMapItem*	m_VCIDMap;				// The mapping table.
	uchar*			m_NextCID;				// For each VOBID, the next CID that should be used.
	ushort			m_NextVID;				// Next VID to be used.

} ;




#endif